CON
  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

  CR = 13
  END_CHR = ">"
  BEGIN_CHR = "<"

  ' Define Pins
  CS = 0        'CLK = CS + 1, MISO = CS + 2, MOSI = CS + 3  -> ADC PINS
  ACC1 = 4
  ACC2 = 5
  ACC3 = 6
  ACC4 = 7
  WIND = 8
  W_STOP = 9
  E_STOP = 10

  XB_IN = 15
  XB_OUT = 16
  XB_BAUD = 9600

  ' Output
  UP = 11
  DOWN = 12
  LEFT = 13
  RIGHT = 14

  ' Offsets
  east_offset = 0
  west_offset = 0
  up_offset = 0
  down_offset = 0

  ' Define Specifics
  Max_Freq_Count = 4095          ' Max number counted from sensor
    
  SampleRate = 70                    ' Time between E/W and U/D sample
  WindSampleRate = 10
  MaxSampleCount = 1000               ' Cloud Buffer
  MaxWindSamples = 50
  ParkWindSpeed = 13             'MPH
  MaxAwakeWind = 6               'MPH

  Min_EW_move = 100                  ' Minimum Pulse count to move East or West
  Min_UD_move = 100                  ' Minimum Pulse count to move Up or Down
  UD_Park_Range = 70
  EW_Park_Range = 70                ' Go to Park mode when EW pulse count equals X
  Park_Timer = 10000              ' Number of loops before Park is initiated
  Wake_Timer = 10000
  
  max_diff = 5

  Wind_Division = 4

  ' Serial Terminal Stuff
  DEBUG_BAUD = 9_600
  DEBUG_HEADING_X = 1          ' Output formatting data
  DEBUG_HEADING_Y = 1
  DEBUG_DYNAMIC_X = DEBUG_HEADING_X + 10
  DEBUG_DYNAMIC_Y = DEBUG_HEADING_Y + 2

VAR
  LONG moveStack[20]
  LONG wifiStack[30]
  LONG EWArray[MaxSampleCount]
  LONG UDArray[MaxSampleCount]
  LONG WindArray[MaxWindSamples]
  LONG East_Count, West_Count, Up_Count, Down_Count  
  LONG EWdiff
  LONG EWdiffavg
  LONG UDdiff
  LONG UDdiffavg
  LONG idx1

  LONG PulseCounterStack[50]
  LONG pulseCount
  LONG pulsewidth

  BYTE WindSpeed
  BYTE WindAVG
  BYTE command
  BYTE EastSwitch
  BYTE WestSwitch
  BYTE Park_Flag
  BYTE threshold_flag
  BYTE moving_flag
  BYTE buffer[256]
  BYTE index
  BYTE last_data[256]
  BYTE hold_position
  BYTE shutdown
  BYTE move_EW_flag
  BYTE move_UD_flag
  BYTE move_to_park
  
OBJ
  adc           : "ADC_MCP3204_SE_SPI"
  Pst           : "Parallax Serial Terminal"
  xb            : "FullDuplexSerial"
  str           : "STRINGS2"

PUB go | EWarray_pointer, UDarray_pointer, Wind_Array_Pointer, aloopcount, arraysum, threshold_count, windloopcount
  moving_flag := 0
  move_to_park := 0
  aloopcount := 0
  windloopcount := 0
  threshold_count := 0
  threshold_flag := 0
  hold_position := 0
  command := 0
  shutdown := 0

  dira[E_STOP]~
  dira[W_STOP]~

  pause(1000)

  Pst.Start(DEBUG_BAUD)                                                         '1 Cog
  pause(500)
  cognew(Get_Wind, @PulseCounterStack)
  pause(500)
  cognew(controlMovement, @moveStack)                                    '1 Cog
  pause(1000)
  cognew(Handle_Wifi, @wifiStack)                                        '2 Cogs
  pause(1000)
  adc.start(CS, 0, @West_Count, @Up_Count, @East_Count, @Down_Count)            '1 Cog
  pause(1000)
  
  Pst.Clear
  Pst.Position(DEBUG_HEADING_X, DEBUG_HEADING_Y)
  Pst.Str(string("Current Position", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("------------------------", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("East  ................", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("West  ................", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("EWdiff ...............", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("EWavg ...............", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("Up  .................", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("Down  ................", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("UDdiff ...............", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("UDavg ...............", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("Park ...............", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("West SW ...............", 13))    
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("East SW ...............", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("Wind ...............", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("Wind AVG ...............", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("Thrshld ...............", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("Moving ...............", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("In Data ...............", 13))
  Pst.PositionX(DEBUG_HEADING_X)
  Pst.Str(string("Command ...............", 13))
  
  repeat
    ReadSwitches  
    Update_Debug_Window

    if East_Count > West_Count
      EWdiff := East_Count - West_Count
    else
      EWdiff := West_Count - East_Count
    
    if Up_Count > Down_Count
      UDdiff := Up_Count - Down_Count
    else
      UDdiff := Down_Count - Up_Count

    aloopcount++
    windloopcount++

    if windloopcount => WindSampleRate
      WindArray[Wind_Array_Pointer] := WindSpeed
      if Wind_Array_Pointer == MaxWindSamples
        Wind_Array_Pointer := 0
      Wind_Array_Pointer++
      WindAVG := getAverage(@WindArray, MaxWindSamples)
      windloopcount := 0
    
    if aloopcount => SampleRate
      if East_Count > Min_EW_move OR West_Count > Min_EW_move
        EWArray[EWarray_pointer] := EWdiff
        EWarray_pointer++
        if EWarray_pointer == MaxSampleCount
          EWarray_pointer := 0
      if Up_Count > Min_UD_move OR Down_Count > Min_UD_move
        UDArray[UDarray_pointer] := UDdiff
        UDarray_pointer++
        if UDarray_pointer == MaxSampleCount
          UDarray_pointer := 0
          EWdiffavg := getAverage(@EWarray, MaxSampleCount)
      UDdiffavg := getAverage(@UDarray, MaxSampleCount)

      aloopcount := 0

    if East_Count > EW_Park_Range OR West_Count > EW_Park_Range OR Up_Count > UD_Park_Range OR Down_Count > UD_Park_Range
      threshold_count++
      if threshold_count > 2000
        threshold_flag := 1
    else
      threshold_flag := 0
      threshold_count := 0
    
PUB Handle_Wifi | char, i
  xb.Start(XB_OUT, XB_IN, %0000, 9_600)
  pause(5000)
  xb.str(string(BEGIN_CHR))
  xb.str(string("Solar Connected!"))
  xb.str(string(END_CHR))
  'xb.tx(10)
  'xb.tx(13)
  index := 0
  i := 0
  repeat
    i++
    index := 0
    if i => 10
      xb.str(string(BEGIN_CHR))
      xb.str(string("PING"))
      xb.str(string(END_CHR))
      'xb.tx(10)
      'xb.tx(13)
      i := 0
    char := xb.rxtime(2000)
    if char == "R"
      repeat until char == "#"
        if char > 31 and char < 126
          buffer[index++] := char
        char := xb.rxtime(1000)
      xb.rxflush
      buffer[index]~
      i := 0
      if(char == "#") AND strsize(@buffer) > 4
        Process_Data
        bytefill(@buffer, 0, 256)
      else
        bytefill(@buffer, 0, 256)
    elseif char > -1
      xb.rxflush
            
    char := -1

PUB Process_Data | send_delay, newcmd
  'idx1 := str.StrStr(@buffer, string("R"), 0)
  'idx2 := str.StrPos(idx1, string(":#"), 0)
  'idx2 := str.SubStr(idx1, 0, idx2)
  'idx3 := str.StrPos(idx2, string("C:"), 0)
  'idx3 := str.SubStr(idx2, idx3 + 2, strsize(idx2) - 1)
  idx1 := str.SubStr(@buffer, 4, 1)
  newcmd := str.StrToDec(idx1)
  pause(500)
  if(newcmd => 0)
    bytemove(@last_data, idx1, 255)
    command := newcmd
     
    case command
      1 :
        move_EW_flag := 1                               ' MOVE EAST
        xb.str(string(BEGIN_CHR))
        xb.str(string("e:1#w:0"))
        xb.str(string(END_CHR))
        'pause(send_delay)
        'xb.tx(10)
        'xb.tx(13)                   
      2 :
        move_EW_flag := 2                               ' MOVE WEST
        xb.str(string(BEGIN_CHR))
        xb.str(string("e:0#w:1"))
        xb.str(string(END_CHR))
        'pause(send_delay)
        'xb.tx(10)
        'xb.tx(13)                   
      3 :
        move_UD_flag := 1                               ' MOVE UP
        xb.str(string(BEGIN_CHR))
        xb.str(string("u:1#d:0"))
        xb.str(string(END_CHR))
        'pause(send_delay)
        'xb.tx(10)
        'xb.tx(13)
      4 :
        move_UD_flag := 2                               ' MOVE DOWN
        xb.str(string(BEGIN_CHR))
        xb.str(string("u:0#d:1"))
        xb.str(string(END_CHR))
        'pause(send_delay)
        'xb.tx(10)
        'xb.tx(13)
      5 :
        move_EW_flag := 0                               ' STOP ALL
        move_UD_flag := 0
        xb.str(string(BEGIN_CHR))
        xb.str(string("e:0#w:0#u:0#d:0"))
        xb.str(string(END_CHR))
        'pause(send_delay)
        'xb.tx(10)
        'xb.tx(13)
      6 :
        xb.str(string(BEGIN_CHR))
        xb.str(string("es:"))
        xb.dec(East_Count)
        xb.str(string("#"))
        xb.str(string("ws:"))
        xb.dec(West_Count)
        xb.str(string("#"))
        xb.str(string("ewdiff:"))
        xb.dec(EWdiff)
        xb.str(string("#"))
        xb.str(string("us:"))
        xb.dec(Up_Count)
        xb.str(string("#"))
        xb.str(string("ds:"))
        xb.dec(Down_Count)
        xb.str(string("#"))
        xb.str(string("uddiff:"))
        xb.dec(UDdiff)
        xb.str(string("#"))
        xb.str(string("wswitch:"))
        xb.dec(WestSwitch)
        xb.str(string("#"))
        xb.str(string("eswitch:"))
        xb.dec(EastSwitch)
        xb.str(string("#"))
        xb.str(string("park:"))
        xb.dec(park_flag)
        xb.str(string("#"))
        xb.str(string("hold:"))
        xb.dec(hold_position)
        xb.str(string("#"))
        xb.str(string("moving:"))
        xb.dec(moving_flag)
        xb.str(string("#"))
        xb.str(string("wind:"))
        xb.dec(WindSpeed)
        xb.str(string("#"))
        xb.str(string("windavg:"))
        xb.dec(WindAVG)
        xb.str(string(END_CHR))
        'pause(send_delay)
        'xb.tx(10)
        'xb.tx(13)
      7 :
        hold_position := 1
        move_EW_flag := 0
        move_UD_flag := 0
        xb.str(string(BEGIN_CHR))
        xb.str(string("Holding Position."))
        xb.str(string(END_CHR))
        'pause(send_delay)
        'xb.tx(10)
        'xb.tx(13)
      8 :
        hold_position := 0
        xb.str(string(BEGIN_CHR))
        xb.str(string("Unholding Position."))
        xb.str(string(END_CHR))
        'pause(send_delay)
        'xb.tx(10)
        'xb.tx(13)
      9 :
        move_to_park := 1
        xb.str(string(BEGIN_CHR))
        xb.str(string("Moving to Park Position"))
        xb.str(string(END_CHR))
        'pause(send_delay)
        'xb.tx(10)
        'xb.tx(13)         
      OTHER :
        move_EW_flag := 0
        move_UD_flag := 0
        move_to_park := 0
        xb.str(string(BEGIN_CHR))
        xb.str(string("Unknown Command"))
        xb.str(string(END_CHR))
        'pause(send_delay)
        'xb.tx(10)
        'xb.tx(13)
    return          
  else
    xb.str(string(BEGIN_CHR))
    xb.str(string("Invalid Data!"))
    xb.str(string(END_CHR))
    'pause(send_delay)
    'xb.tx(10)
    'xb.tx(13)
    return
    
PUB controlMovement | park_loopcount, wake_loopcount, moving_east, moving_west, moving_up, moving_down
  dira[LEFT]~~
  dira[RIGHT]~~
  dira[UP]~~
  dira[DOWN]~~

  outa[LEFT] := 0
  outa[RIGHT] := 0
  outa[UP] := 0
  outa[DOWN] := 0

  moving_east := 0
  moving_west := 0
  moving_up := 0
  moving_down := 0

  wake_loopcount := 0
  park_loopcount := 0
  
  repeat
    if EastSwitch == 1 OR WestSwitch == 1
      EWoff

    if WindSpeed => ParkWindSpeed AND Park_Flag == 0
      shutdown := 1
      command := 0
      Park

    if shutdown == 1 AND WindAVG =< MaxAwakeWind
      shutdown := 0

    if command > 0
      case move_EW_flag
        1 : goEast
        2 : goWest
        OTHER :
          EWoff

      case move_UD_flag
        1 : goUP
        2 : goDown
        OTHER :
          UDoff

      if move_to_park == 1
        move_to_park := 0
        Park

    if hold_position == 0 AND shutdown == 0
      if EWdiffavg => max_diff
        if East_Count > West_Count AND moving_west == 0
          goEast
          moving_east := 1  
        elseif West_Count > East_Count AND moving_east == 0
          goWest
          moving_west := 1
        else
          EWoff
          moving_east := 0
          moving_west := 0
                  
      if UDdiffavg => max_diff
        if Up_Count > Down_Count AND moving_down == 0
          goUp
          moving_up := 1
        elseif Down_Count > Up_Count AND moving_up == 0
          goDown
          moving_down := 1
        else
          UDoff
          moving_down := 0
          moving_up := 0
          
      if East_Count < EW_Park_Range AND West_Count < EW_Park_Range AND Up_Count < UD_Park_Range AND Down_Count < UD_Park_Range AND Park_Flag == 0
        park_loopcount++
        wake_loopcount := 0
        if park_loopcount > Park_Timer
          park_loopcount := 0
          Park
      else                                                                    
        if East_Count > EW_Park_Range OR West_Count > EW_Park_Range OR Up_Count > UD_Park_Range OR Down_Count > UD_Park_Range
          wake_loopcount++
          if Park_Flag == 1 AND moving_flag == 0 AND threshold_flag == 1 AND wake_loopcount > Wake_Timer
            Wake
            wake_loopcount := 0 
        park_loopcount := 0

PUB getAverage(the_array, sample_count) | i, arraysum
    arraysum := 0
    repeat i from 0 to sample_count
      arraysum += long[the_array][i]
    return arraysum / sample_count
     
PUB goEast
  outa[RIGHT] := 0
  if EastSwitch == 0
    moving_flag := 1
    outa[LEFT] := 1
  else
    outa[LEFT] := 0
  return

PUB goWest
  outa[LEFT] := 0
  if WestSwitch == 0
    moving_flag := 1
    outa[RIGHT] := 1
  else
    outa[RIGHT] := 0
  return

PUB goUp
  outa[DOWN] := 0
  moving_flag := 1
  outa[UP] := 1
  return
  
PUB goDown
  outa[UP] := 0
  moving_flag := 1
  outa[DOWN] := 1
  return

PUB EWoff
  outa[LEFT] := 0
  outa[RIGHT] := 0
  if move_UD_flag == 0
    moving_flag := 0
  reset_EW_avg
  return

PUB UDoff
  outa[UP] := 0
  outa[DOWN] := 0
  if move_EW_flag == 0
    moving_flag := 0
  reset_UD_avg
  return

PUB reset_EW_avg
  longfill(@EWarray, 0, MaxSampleCount)
  EWdiffavg := 0
  return

PUB reset_UD_avg
  longfill(@UDarray, 0, MaxSampleCount)
  UDdiffavg := 0
  'pause(1000)
  return
  
PUB ReadSwitches | i
  if ina[W_STOP] == 1
    WestSwitch := 0
  else
    WestSwitch := 1
    
  if ina[E_STOP] == 1
    EastSwitch := 0
  else
    EastSwitch := 1
    
  return

PUB Park | i
  Park_Flag := 1
  threshold_flag := 0
  UDoff
  EWoff
  pause(100)
  i := 0
  'goUp

  repeat until WestSwitch == 1 or threshold_flag == 1
    if command > 0 AND command < 6
      quit
    goWest
  EWoff
  pause(2000)
  
  repeat until i == 1199 or threshold_flag == 1
    if command > 0 AND command < 6
      quit
   goEast
    pause(10)
    i++
  EWoff

  pause(2000)
  
  i := 0
  repeat until i == 30000 or threshold_flag == 1
    if command > 0 AND command < 6
      quit
    goUp
    pause(10)
    i++
  UDoff  
  return

PUB Wake | i
  Park_Flag := 0
  UDoff
  EWoff
  pause(100)

  repeat until EastSwitch == 1
    if command > 0 AND command < 6
      quit
    goEast
    pause(10)
  EWoff

  pause(2000)
  
  i := 0
  repeat until i == 20000 or threshold_flag == 1 and i => 5000
    if command > 0 AND command < 6
      quit
    goDown
    pause(10)
    i++
  UDoff
  return

PUB Update_Debug_Window
    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y)
    Pst.Dec(East_Count)
    Pst.Str(string("..."))
    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 1)
    Pst.Dec(West_Count)
    Pst.Str(string("..."))
    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 2)
    Pst.Dec(EWdiff)
    Pst.Str(string("..."))
    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 3)
    Pst.Dec(EWdiffavg)
    Pst.Str(string("..."))
    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 4)
    Pst.Dec(Up_Count)
    Pst.Str(string("..."))
    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 5)
    Pst.Dec(Down_Count)
    Pst.Str(string("..."))
    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 6)
    Pst.Dec(UDdiff)
    Pst.Str(string("..."))
    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 7)
    Pst.Dec(UDdiffavg)
    Pst.Str(string("..."))

    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 8)
    Pst.Dec(Park_Flag)
    Pst.Str(string("..."))
    
    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 9)
    Pst.Dec(WestSwitch)
    Pst.Str(string("..."))

    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 10)
    Pst.Dec(EastSwitch)
    Pst.Str(string("..."))

    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 11)
    Pst.Dec(WindSpeed)
    Pst.Str(string("..."))

    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 12)
    Pst.Dec(WindAVG)
    Pst.Str(string("..."))

    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 13)
    Pst.Dec(threshold_flag)
    Pst.Str(string("..."))

    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 14)
    Pst.Dec(moving_flag)
    Pst.Str(string("..."))

    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 15)
    Pst.Str(@last_data)
    Pst.Str(string("..."))

    Pst.Position(DEBUG_DYNAMIC_X, DEBUG_DYNAMIC_Y + 16)
    Pst.dec(command)
    Pst.Str(string("..."))
    return

PUB Get_Wind | updated, cnt1, cnt2, count1
  ' 65825 counts per 1 second
  count1 := 0
  pulsewidth := 60000 ' Fill to prevent 0 division.
  WindSpeed := 0
  repeat
    updated := 0
    repeat while ina[WIND] == 1
      if updated == 0
        cnt2 := cnt
        pulseCount++
        updated := 1
      waitcnt(clkfreq / 1000 * 5 + cnt)
      if count1 > 0
        pulsewidth := (cnt2 - cnt1) / (clkfreq / 10000) 
      count1 := 0
    if updated == 1
      cnt1 := cnt
    if(pulsewidth > 0)
      WindSpeed := (60000 / pulsewidth) '* Wind_Division

    if(WindSpeed < 0)
      WindSpeed := WindSpeed * -1
      
    count1++
    
PUB pause(Duration)
  if(Duration > 0)  
    waitcnt((clkfreq / 1_000 * Duration) + cnt)
  return